# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import importlib

from piniverse.common.responsibilities.responsibility import Responsibility
from piniverse.common.exceptions.pin_exception import PinException


class ModuleFinder(Responsibility):
    """ Module Finder

       Attributes:
            key (str): package identifier
            value (object): package or package name
            successor (Responsibility): responsibility's successor
    """

    def __init__(self, successor: 'Responsibility' = None):
        super(ModuleFinder, self).__init__(successor=successor)

    def _execute(self, content):
        if 'package' not in content:
            raise PinException('Package is undefined.\
                Expected input format { \'package\': [package name] }')

        package = ModuleFinder.parse(content['package'])
        path = importlib.import_module(package).__file__.replace('__init__.py', '')

        targets = []
        for root, dirs, files in os.walk(path):
            for _file in files:
                prefix = root.replace(os.sep, '.')
                index = prefix.find(package)

                if prefix.endswith('.'):
                    prefix = prefix[index:][:-1]
                else:
                    prefix = prefix[index:]

                if _file.endswith('.py'):
                    targets.append('{}.{}'.format(prefix, _file.replace('.py', '')))

        return self._continue_or_return({
            'modules': targets
        })

    @staticmethod
    def parse(package):
        if isinstance(package, str):
            return package
        else:
            return package.__name__



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from piniverse.core.pin import PinOrchestrator
from piniverse.core.module_finder import ModuleFinder
from piniverse.core.pin_finder import PinFinder
from piniverse.common.responsibilities.responsibility import ResponsibilityUtils
import tests.integration_tests.test_core.my_package as my_package
from tests.integration_tests.test_core.my_package.my_file import foo, another_foo
from tests.integration_tests.test_core.my_package.my_directory.my_file import yet_another_foo


def test_pin_orchestrator(capsys):
    pin_orchestrator = PinOrchestrator(
        pin_responsibilities=ResponsibilityUtils.build(ModuleFinder(), PinFinder()))

    pin_orchestrator.plan(my_package)

    parents = pin_orchestrator.pin_graph.parents
    assert '1' == parents['1']
    assert '1' == parents['2']
    assert '3' == parents['3']

    dag = pin_orchestrator.pin_graph.dag
    pin_nodes = dag
    tasks = list(map(lambda a: a.task, pin_nodes))
    assert '1' in tasks
    assert '2' in tasks
    assert '3' in tasks
    assert '1' == tasks[0]
    assert '2' == tasks[1]
    assert '3' == tasks[2]

    functions = list(map(lambda a: a.function, pin_nodes))
    assert foo in functions
    assert another_foo in functions
    assert yet_another_foo in functions

    pin_orchestrator.apply()
    captured = capsys.readouterr()
    assert 'foo, foo done, another foo\nNone\n' == captured.out
    assert '' == captured.err



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import inspect

from piniverse.core.pin_node import PinNode, PinGraph
from piniverse.common.exceptions.pin_exception import PinException
from tests.unit_tests.test_core.my_package.my_file import foo, another_foo


def test_pin_node():
    pin_node_1 = PinNode('1', '2', arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=foo)
    assert '1' == pin_node_1.task
    assert '2' == pin_node_1.toward
    assert '3' == pin_node_1.arguments['args'][0]
    assert '5' == pin_node_1.arguments['kwargs']['4']
    assert foo == pin_node_1.function

    argspecs_1 = pin_node_1.argspecs()
    assert [] == argspecs_1['args']
    assert not argspecs_1['varargs']
    assert not argspecs_1['varkw']
    assert not argspecs_1['defaults']
    assert [] == argspecs_1['kwonlyargs']
    assert not argspecs_1['kwonlydefaults']
    assert '{}' == argspecs_1['annotations']

    pin_node_2 = PinNode('2', '2', arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=another_foo)
    assert '2' == pin_node_2.task
    assert '2' == pin_node_2.toward
    assert '3' == pin_node_2.arguments['args'][0]
    assert '5' == pin_node_2.arguments['kwargs']['4']
    assert another_foo == pin_node_2.function

    argspecs = PinNode.parse(inspect.getfullargspec(foo))
    assert [] == argspecs['args']
    assert not argspecs['varargs']
    assert not argspecs['varkw']
    assert not argspecs['defaults']
    assert [] == argspecs['kwonlyargs']
    assert not argspecs['kwonlydefaults']
    assert '{}' == argspecs['annotations']

    argspecs_2 = pin_node_2.argspecs()
    assert [] == argspecs_2['args']
    assert not argspecs_2['varargs']
    assert not argspecs_2['varkw']
    assert not argspecs_2['defaults']
    assert [] == argspecs_2['kwonlyargs']
    assert not argspecs_2['kwonlydefaults']
    assert '{}' == argspecs_2['annotations']


def test_pin_graph():
    parents = {'1': '1', '2': '1', '3': '2', '4': '4', '5': '4'}
    assert '1' == PinGraph.find(parents, '1')
    assert '1' == PinGraph.find(parents, '2')
    assert '1' == PinGraph.find(parents, '3')
    assert '4' == PinGraph.find(parents, '4')
    assert '4' == PinGraph.find(parents, '5')

    pin_node_1 = PinNode('1', '2', arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=foo)
    pin_node_2 = PinNode('2', '1', arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=another_foo)
    pin_nodes = [pin_node_1, pin_node_2]
    with pytest.raises(PinException) as e:
        PinGraph.union(pin_nodes)
    assert 'Found a cyclic dependency: 1, 2' == str(e.value) or 'Found a cyclic dependency: 2, 1' == str(e.value)

    pin_node_1 = PinNode('1', '2', arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=foo)
    pin_node_2 = PinNode('2', None, arguments={'args': ['3'], 'kwargs': {'4': '5'}}, function=another_foo)
    pin_nodes = [pin_node_1, pin_node_2]
    parents = PinGraph.union(pin_nodes)
    assert '1' == parents['1']
    assert '1' == parents['2']

    dag = PinGraph.topological_sort(pin_nodes)
    assert foo == next(
        filter(lambda pin_node: pin_node.task == '1', dag)).function
    assert another_foo == next(
        filter(lambda pin_node: pin_node.task == '2', dag)).function

    dag = PinGraph(pin_nodes).dag
    assert foo == next(
        filter(lambda pin_node: pin_node.task == '1', dag)).function
    assert another_foo == next(
        filter(lambda pin_node: pin_node.task == '2', dag)).function



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect

from piniverse.common.exceptions.pin_exception import PinException


class PinNode:
    """ Base class for pin nodes

        task (str): task
        toward (str): toward
        arguments (dict): arguments of wrapped function
        function: wrapped function
    """

    def __init__(self, task: str, toward: str, arguments: dict, function = None):
        self._task = task
        self._toward = toward
        self._arguments = arguments
        self._function = function

    @property
    def task(self):
        return self._task

    @property
    def toward(self):
        return self._toward

    @property
    def arguments(self):
        return self._arguments

    @property
    def function(self):
        return self._function

    def argspecs(self):
        return self.parse(inspect.getfullargspec(self._function))

    @staticmethod
    def parse(argspecs):
        return {
            'args': argspecs.args,
            'varargs': argspecs.varargs,
            'varkw': argspecs.varkw,
            'defaults': argspecs.defaults,
            'kwonlyargs': argspecs.kwonlyargs,
            'kwonlydefaults': argspecs.kwonlydefaults,
            'annotations': str(argspecs.annotations)
        }


class PinGraph:
    """ Base class for pin graph

        Attributes:
            parents (dict): parents to respective tasks
            dag (list): direct acyclic graph of pin nodes
    """

    def __init__(self, pin_nodes: list):
        self._parents = PinGraph.union(pin_nodes)
        self._dag = PinGraph.topological_sort(pin_nodes)

    @property
    def parents(self):
        return self._parents

    @property
    def dag(self):
        return self._dag

    @staticmethod
    def topological_sort(pin_nodes: list) -> list:
        inwards = {}
        outwards = {}

        for pin_node in pin_nodes:
            if pin_node.task not in inwards or pin_node.task not in outwards:
                inwards[pin_node.task] = 0
                outwards[pin_node.task] = 0

        for pin_node in pin_nodes:
            if pin_node.toward:
                outwards[pin_node.task] += 1
                inwards[pin_node.toward] += 1

        actives = []
        orphans = []
        for pin_node in pin_nodes:
            if inwards[pin_node.task] == 0 and outwards[pin_node.task] == 0:
                orphans.append(pin_node)
            else:
                actives.append(pin_node)

        ref = {}
        for pin_node in actives:
            if pin_node.task not in ref:
                ref[pin_node.task] = pin_node

        graph = {}
        degrees = {}
        for pin_node in actives:
            task = pin_node.task
            toward = pin_node.toward

            if not toward:
                continue

            if task not in graph:
                graph[task] = []

            if task not in degrees:
                degrees[task] = 0

            if toward not in graph:
                graph[toward] = []

            if toward not in degrees:
                degrees[toward] = 0

            graph[task].append(toward)
            degrees[toward] += 1

        q = []
        for k, v in degrees.items():
            if v == 0:
                q.append(k)

        order = []
        while q:
            size = len(q)
            for i in range(size):
                cur = q.pop(0)
                order.append(ref[cur])

                for _next in graph[cur]:
                    degrees[_next] -= 1
                    if degrees[_next] == 0:
                        q.append(_next)

        return order + orphans

    @staticmethod
    def union(pin_nodes: list) -> list:
        parents = {}

        for pin_node in pin_nodes:
            parents[pin_node.task] = pin_node.task
            if pin_node.toward is not None:
                parents[pin_node.toward] = pin_node.toward

        for pin_node in pin_nodes:
            if pin_node.toward:
                left = PinGraph.find(parents, pin_node.task)
                right = PinGraph.find(parents, pin_node.toward)

                if left != right:
                    parents[right] = left
                else:
                    raise PinException(
                        'Found a cyclic dependency: {}, {}'
                        .format(pin_node.task, pin_node.toward))

        return parents

    @staticmethod
    def find(parents: dict, cur: int) -> int:
        if parents[cur] != cur:
            parents[cur] = PinGraph.find(parents, parents[cur])
        return parents[cur]



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from piniverse.core.pin import Pin, Pinned
from piniverse.core.tags import PinKeys, PinValues
from piniverse.common.exceptions.pin_exception import PinException


def test_pin():
    pin = Pin('1', '2', {'args': ['3'], 'kwargs': {'4': '5'}})
    assert '1' == pin.task
    assert '2' == pin.toward
    assert '3' == pin.arguments['args'][0]
    assert '5' == pin.arguments['kwargs']['4']


def test_pinned():
    with pytest.raises(PinException) as e:
        Pinned()
    assert 'Task is undefined' == str(e.value)

    with pytest.raises(PinException) as e:
        Pinned(toward='1')
    assert 'Task is undefined' == str(e.value)

    pinned = Pinned(task='1', toward='2')
    assert '1' == pinned.task
    assert '2' == pinned.toward
    assert 0 == len(pinned.arguments['args'])
    assert 0 == len(pinned.arguments['kwargs'])

    @Pinned(
        task='1',
        toward='2',
        arguments={'args': ['3'], 'kwargs': {'4': '5'}}
    )
    def foo(a, b=''):
        pass

    assert '1' == getattr(foo, PinKeys.TASK, '')
    assert '2' == getattr(foo, PinKeys.TOWARD, '')
    assert PinValues.PINNED == getattr(foo, PinKeys.TAG, '')
    assert '3' == getattr(foo, PinKeys.ARGUMENTS, None)['args'][0]
    assert '5' == getattr(foo, PinKeys.ARGUMENTS, None)['kwargs']['4']


def tes_pin_orchestrator():
    pass



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from piniverse.core.pin_finder import PinFinder
from tests.unit_tests.test_core import my_package


def test_pin_finder():
    pin_finder = PinFinder()
    assert 'default' == pin_finder.key
    assert not pin_finder.value
    assert not pin_finder.successor
    assert 0 == len(pin_finder.pin_nodes)

    pin_nodes = pin_finder.execute(
        {'modules': ['tests.unit_tests.test_core.my_package.my_file']}
    )['pin_nodes']
    assert 2 == len(pin_nodes)
    assert my_package.my_file.foo == next(
        filter(lambda pin_node: pin_node.task == '1', pin_nodes)).function
    assert my_package.my_file.another_foo == next(
        filter(lambda pin_node: pin_node.task == '2', pin_nodes)).function



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class PinStore():
    """ Pin Store

        Attributes:
            state (dict): state of the store
    """

    def __init__(self):
        self._state = {
            'returns': {}
        }

    def push(self, key: str = None, content=None) -> None:
        if key:
            self._state[key] = content

    def pull(self, key: str = None):
        if key:
            return self._state.get(key, None)

    def rpush(self, task: str = None, content=None) -> None:
        if task:
            self._state['returns'][task] = content

    def rpull(self, task: str = None):
        if task:
            return self._state['returns'].get(task, None)



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from piniverse.core.pin_node import PinGraph
from piniverse.core.pin_store import PinStore
from piniverse.core.tags import PinKeys, PinValues
from piniverse.common.exceptions.pin_exception import PinException


class Pin:
    """ Pin

        Attributes:
            task (int): task number
            toward (int): toward next task number
            arguments (dict): arguments for wrapped function
    """

    def __init__(self, task: str, toward: str, arguments: dict):
        self._task = task
        self._toward = toward
        self._arguments = arguments

    @property
    def task(self):
        return self._task

    @property
    def toward(self):
        return self._toward

    @property
    def arguments(self):
        return self._arguments


class PinDecorator(Pin):
    """ Pin Decorator """

    def __init__(self, task: str = None, toward: str = None, arguments: dict = {'args': [], 'kwargs': {}}):
        if not task:
            raise PinException('Task is undefined')

        if 'args' not in arguments:
            arguments['args'] = []

        if 'kwargs' not in arguments:
            arguments['kwargs'] = {}

        super(PinDecorator, self).__init__(
            task,
            toward if task != toward else None,
            arguments)

    def __call__(self, child):
        setattr(child, PinKeys.TASK, self._task)
        setattr(child, PinKeys.TOWARD, self._toward)
        setattr(child, PinKeys.ARGUMENTS, self._arguments)
        setattr(child, PinKeys.TAG, PinValues.PINNED)
        return child


Pinned = PinDecorator


class PinOrchestrator():
    """ Pin Orchestrator

        Attributes:
            title (str): title of the orchestrator
            responsibilities (Responsibility): list of responsibilities
            pin_graph (PinGraph): graph of pin nodes
    """

    def __init__(self, title: str = 'master', pin_responsibilities=None):
        self._title = title
        self._pin_responsibilities = pin_responsibilities
        self._pin_graph = None

    @property
    def pin_graph(self):
        return self._pin_graph

    def apply(self):
        if not self._pin_graph:
            raise PinException('Pin Graph is undefined. Plan before apply')

        stores = {}
        for k, v in self._pin_graph.parents.items():
            if k == v:
                stores[k] = PinStore()

        for pin_node in self._pin_graph.dag:
            task = pin_node.task
            args = pin_node.arguments
            _callable = pin_node.function

            try:
                parent = self._pin_graph.parents[task]
                store = stores[parent]

                kwargs = {
                    **args['kwargs'],
                    'store': store
                }

                result = _callable(*args['args'], **kwargs)
                store.rpush(task=task, content=result)
            except Exception as e:
                raise PinException(
                    'Failed to execute callable {} due to {}'.format(
                        _callable.__name__,
                        str(e)))

    def plan(self, package):
        pin_nodes = self._pin_responsibilities.execute(
            {'package': package}
        )['pin_nodes']

        self._pin_graph = PinGraph(pin_nodes)

    def orchestrate(self):
        pass



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib
import inspect

from piniverse.core.pin_node import PinNode
from piniverse.core.tags import PinKeys, PinValues
from piniverse.common.responsibilities.responsibility import Responsibility
from piniverse.common.visitors.visitor import Visitor


class PinFinder(Responsibility, Visitor):
    """ Pin Finder

        Attributes:
            pin_nodes (list): list of pin nodes
            successor (Responsibility): succeeding responsibility
    """

    def __init__(self, successor: 'Responsibility' = None):
        super(PinFinder, self).__init__(successor=successor)
        self._pin_nodes = []

    @property
    def pin_nodes(self):
        return self._pin_nodes

    def _execute(self, content):
        modules = content['modules']

        for module in modules:
            imported = importlib.import_module(module)
            members = inspect.getmembers(imported, predicate=inspect.isfunction)

            for member in members:
                function = member[1]
                self.visit(function)

        return self._continue_or_return({
            'pin_nodes': self._pin_nodes
        })

    def visit(self, visitable):
        if getattr(visitable, PinKeys.TAG, '') == PinValues.PINNED:
            pin_node = PinNode(
                getattr(visitable, PinKeys.TASK),
                getattr(visitable, PinKeys.TOWARD),
                getattr(visitable, PinKeys.ARGUMENTS),
                visitable)

            self._pin_nodes.append(pin_node)



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Responsibility:
    """ Abstract base class for responsibilities

        To derive this class, it is expected to ovveride the '_execute' method.

        Attributes:
            key (str): responsibility's key
            value (object): responsibility's value
            successor (Responsibility): responsibility's successor
    """

    def __init__(self, key: str = 'default', value = None, successor: 'Responsibility' = None):
        self._key = key
        self._value = value
        self._successor = successor

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> str:
        return self._value

    @property
    def successor(self) -> 'Responsibility':
        return self._successor

    def execute(self, content: dict):
        return self._execute(content)

    def _execute(self, content: dict):
        pass

    def _continue_or_return(self, content: dict) -> dict:
        if self._successor:
            return self._successor.execute(content)
        else:
            return content


class ResponsibilityUtils:
    """ Utility functions for the Responsiblity class """

    @staticmethod
    def build(*responsibilities: list) -> 'Responsibility':
        head = Responsibility()
        pre = head

        for responsibility in responsibilities:
            pre._successor = responsibility
            pre = pre._successor

        return head._successor

    @staticmethod
    def pretty_print(responsibility):
        cur = responsibility
        print('\n----- Responsibilities -----\n')

        while cur:
            print('[key: {}, value: {}] -> \n'.format(cur.key, cur.value))
            cur = cur.successor

        print('null\n')
        print('\n----------------------------\n')



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from piniverse.core.pin import PinOrchestrator
from piniverse.core.module_finder import ModuleFinder
from piniverse.core.pin_finder import PinFinder
from piniverse.core.pin_drawer import PinDrawer
from piniverse.common.responsibilities.responsibility \
    import ResponsibilityUtils

# Pin Orchestrator
pin_orchestrator = None


def apply():
    global pin_orchestrator
    pin_orchestrator.apply()


def plan(package, plan_view: bool = False):
    global pin_orchestrator
    pin_orchestrator = PinOrchestrator(
        pin_responsibilities=ResponsibilityUtils.build(
            ModuleFinder(), PinFinder()
        )
    )

    pin_orchestrator.plan(package)

    if plan_view:
        PinDrawer(pin_orchestrator.dag).illustrate()


def orchestrate():
    pass



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from piniverse.common.responsibilities.responsibility import Responsibility


def test_responsibility():
    a = Responsibility(key='a', value='1')
    assert 'a' == a.key
    assert '1' == a.value
    assert not a.successor

    c = Responsibility(key='c', value='3')
    b = Responsibility(key='b', value='2', successor=c)
    assert 'b' == b.key
    assert '2' == b.value
    assert c == b.successor
    assert 'c' == c.key
    assert '3' == c.value
    assert not c.successor


def test_responsibility_utils(capsys):
    a = Responsibility(key='a', value='1')
    assert 'a' == a.key
    assert '1' == a.value
    assert not a.successor

    b = Responsibility(key='b', value='2')
    assert 'b' == b.key
    assert '2' == b.value
    assert not b.successor

    # responsibilities = ResponsibilityUtils.build(a, b)
    # assert b == a.successor
    # assert not b.successor

    # ResponsibilityUtils.pretty_print(responsibilities)
    # captured = capsys.readouterr()
    # assert captured.out == '\n----- Responsibilities -----
    # \n[key: a, value: 1] -> \n[key: b, value: 2] -> \nnull
    # \n----------------------------\n'
    # assert captured.err == ''



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from piniverse import Pinned


@Pinned(
    task='1',
    toward='2',
    arguments={
        'args': ['foo']
    }
)
def foo(content, **kwargs):
    kwargs['store'].push(key='k', content=content)
    return 'foo done'


@Pinned(
    task='2',
    arguments={
        'kwargs': {'content': 'another foo'}
    }
)
def another_foo(content = '', **kwargs):
    full_content = '{}, {}, {}'.format(
        kwargs['store'].pull(key='k'),
        kwargs['store'].rpull(task='1'),
        content
    )

    print(full_content)



# The MIT License (MIT)
#
# Copyright (c) 2019 Henry Zhao
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup, find_packages


setup(
    name = 'piniverse',
    version = '0.0.1',  
    license='MIT',
    description = 'Pin your functions!',
    author = 'Henry Zhao',
    author_email = 'hzhao1016@gmail.com',
    url = 'https://github.com/hzhao19/piniverse',
    keywords = ['pin', 'pinned', 'dag'],
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'matplotlib==3.1.1',
        'networkx==2.4',
        'pytest==5.2.1'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ],
    extras_require={
        'dev': [''],
        'test': ['coverage', 'pytest'],
    }
)
